/*
QDM.cpp
Written by Walter O. Krawec

Copyright (c) 2013 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "QDM.h"

#ifdef WIN32
// Define _kbhit()
#include <conio.h>
#endif

namespace qdm
{
	void QDM::setup(void* netRTGraph)
	{
#ifdef USE_VISUAL_DEBUG
		qp.registerCommand(new CommandGraph2D((wok::NetRTGraphClient*)netRTGraph, this));
#endif

		qp.registerCommand(new CommandDebug(this));
	}
	bool QDM::saveToFile(const std::string& filename)
	{
		std::ofstream f;
		f.open(filename.c_str(), std::ios_base::binary);
		if(!f.is_open())
			return false;
		int FILE_VERSION=10;
		f.write((char*)&FILE_VERSION, sizeof(int));
		int sensorSize = (int)sensors.size();
		f.write((char*)&sensorSize, sizeof(int));

		for(int i=0; i<sensorSize; ++i)
		{
			int stringSize = sensors[i].name.size();
			f.write((char*)&stringSize, sizeof(int));
			for(int j=0; j<stringSize; ++j)
				f.put(sensors[i].name[j]);

			int vectorSize = sensors[i].data.chi.size();
			f.write((char*)&vectorSize, sizeof(int));
			for(int j=0; j<vectorSize; ++j)
			{
				double value = sensors[i].data.chi[j];
				f.write((char*)&value, sizeof(double));
			}

			vectorSize = sensors[i].data.phi.size();
			f.write((char*)&vectorSize, sizeof(int));
			for(int j=0; j<vectorSize; ++j)
			{
				double value = sensors[i].data.phi[j];
				f.write((char*)&value, sizeof(double));
			}

			vectorSize = sensors[i].data.psi.size();
			f.write((char*)&vectorSize, sizeof(int));
			for(int j=0; j<vectorSize; ++j)
			{
				double value = sensors[i].data.psi[j];
				f.write((char*)&value, sizeof(double));
			}
		}

		f.close();
		return true;
	}
	bool QDM::loadFromFile(const std::string& filename)
	{
		std::ifstream f;
		f.open(filename.c_str(), std::ios_base::binary);
		if(!f.is_open())
			return false;

		int version, sensorSize;
		f.read((char*)&version, sizeof(int));
		if(version != 10) return false;
		f.read((char*)&sensorSize, sizeof(int));
		
		for(int i=0; i<sensorSize; ++i)
		{
			Sensor sensor;
			int stringSize;
			f.read((char*)&stringSize, sizeof(int));
			sensor.name.resize(stringSize, ' ');
			for(int j=0; j<stringSize; ++j)
				sensor.name[j] = f.get();

			int vectorSize;
			f.read((char*)&vectorSize, sizeof(int));
			for(int j=0; j<vectorSize; ++j)
			{
				double value;
				f.read((char*)&value, sizeof(double));
				sensor.data.chi.push_back(value);
			}

			f.read((char*)&vectorSize, sizeof(int));
			for(int j=0; j<vectorSize; ++j)
			{
				double value;
				f.read((char*)&value, sizeof(double));
				sensor.data.phi.push_back(value);
			}

			f.read((char*)&vectorSize, sizeof(int));
			for(int j=0; j<vectorSize; ++j)
			{
				double value;
				f.read((char*)&value, sizeof(double));
				sensor.data.psi.push_back(value);
			}

			sensor.data.op.create(sensor.data.chi.size());
			sensor.data.refreshOp();
			sensors.push_back(sensor);
		}

		f.close();

		std::cout << "Sensor size is " << sensors.size() << "\n";
		return true;
	}

	bool QDM::loadFromFiles(const std::vector < std::string >& filenames)
	{
		sensors.clear();
		algebra::mat temp, temp2;
		for(unsigned int i=0; i<filenames.size(); ++i)
		{
			std::ifstream f;
			f.open(filenames[i].c_str(), std::ios_base::binary);
			if(!f.is_open())
				return false;
			int version, sensorSize;
			f.read((char*)&version, sizeof(int));
			if(version != 10) return false;
			f.read((char*)&sensorSize, sizeof(int));
			std::cout << "Load: sensorSize = " << sensorSize << "\n";
			for(int i=0; i<sensorSize; ++i)
			{
				Sensor sensor;
				int stringSize;
				f.read((char*)&stringSize, sizeof(int));
				sensor.name.resize(stringSize, ' ');
				for(int j=0; j<stringSize; ++j)
					sensor.name[j] = f.get();

				int vectorSize;
				f.read((char*)&vectorSize, sizeof(int));
				for(int j=0; j<vectorSize; ++j)
				{
					double value;
					f.read((char*)&value, sizeof(double));
					sensor.data.chi.push_back(value);
					if(j < 5)
						std::cout << "Load: value = " << value << "\n";
				}

				f.read((char*)&vectorSize, sizeof(int));
				for(int j=0; j<vectorSize; ++j)
				{
					double value;
					f.read((char*)&value, sizeof(double));
					sensor.data.phi.push_back(value);
				}

				f.read((char*)&vectorSize, sizeof(int));
				for(int j=0; j<vectorSize; ++j)
				{
					double value;
					f.read((char*)&value, sizeof(double));
					sensor.data.psi.push_back(value);
				}

				sensor.data.op.create(sensor.data.chi.size());
				sensor.data.refreshOp();
				temp.create(actionSize);
				temp.ident();
				sensor.data.op.tensor(&temp, &temp2);
				sensor.data.op = temp2;
				sensors.push_back(sensor);
			}

			f.close();
		}
		return true;
	}

	bool QDM::addActionSpace(ActionSpace& A)
	{
		QProg::Object output;
		int maxGridSize = A.gridSize;

		// create Delta space:
		std::stringstream str;
		//str << "new space Delta." << actions.size() << " (integer " << A.points.size() << ")";	// used to be size()+1
		str << "begin ";
		int localDeltaSize = 1;
		for(int i=0; i<A.getDim(); ++i)
		{
			str << "(new space Delta." << actions.size() << "." << i << " (integer " << A.movementMap.size() << ") ) "; //2*A.freedomOfMovement+1 << ") ) ";
			localDeltaSize *= A.movementMap.size();
			for(int j=0; j<(int)A.movementMap.size(); ++j)
			{
				str << "(new integer Delta." << actions.size() << "." << i << ".dir." << j << ") ";
				int dir = 0;
				dir = A.movementMap[j].coord[0];
				str << "(set Delta." << actions.size() << "." << i << ".dir." << j << " (integer " << dir << ") ) ";
			}
		}
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		// create Action space:
		//str = std::stringstream();
		str.str("");
		str << "begin ";
		int sum = 0;
		/*std::vector <int> points;
		int spaceSize = 1;
		for(unsigned int i=0; i<A.getDim(); ++i)
		{
			spaceSize *= maxGridSize;
			points.push_back(0);
		}*/

		int localActionSize = 1;
		for(int i=0; i<A.getDim(); ++i)
		{
			str << "(new space Action." << actions.size() << "." << i << " (integer " << maxGridSize << ") ) ";
			str << "(new integer Action." << actions.size() << "." << i << ".result) ";
			localActionSize *= maxGridSize;
		}
		str << "(new integer Action." << actions.size() << ".dim) ";
		str << "(set Action." << actions.size() << ".dim (integer " << A.getDim() << ") ) ";
		str << "(new integer Action." << actions.size() << ".size) ";
		str << "(set Action." << actions.size() << ".size (integer " << localActionSize << ") ) ";
		str << "(new integer Delta." << actions.size() << ".size) ";
		str << "(set Delta." << actions.size() << ".size (integer " << localDeltaSize << ") )";
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		//str = std::stringstream();
		//A.actionSize = sum;
		//str << "new space Action." << actions.size() << " (integer " << sum << ")";
		//qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		//A.gridSize = maxGridSize;
		actions.push_back(A);
		return true;
	}

	bool QDM::finalizeSpace(int _memorySize)
	{
		memorySize = _memorySize;
		QProg::Object output;
		std::stringstream str;
		int spaceSize = 0;//(int)actions.size();
		for(unsigned int i=0; i<actions.size(); ++i)
			spaceSize += 2*(int)actions[i].getDim();

		str << "new space Memory (integer " << memorySize << ")";
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);
		
		str.str("");

		str << "new system QDM (integer " << spaceSize+1 << ") Memory ";

		deltaSize = 1;
		for(unsigned int i=0; i<actions.size(); ++i)
		{
			for(int j=0; j<actions[i].getDim(); ++j)
			{
				//deltaSize *= (2*actions[i].freedomOfMovement+1);
				deltaSize *= actions[i].movementMap.size();
				str << "Delta." << i << "." << j << " ";
			}
		}

		//deltaSize = 1;
		actionSize = 1;
		for(unsigned int i=0; i<actions.size(); ++i)
		{
			//deltaSize *= (actions[i].getDim()+1);//(actions[i].points.size());//(actions[i].points.size()+1);
			//deltaSize *= (2*actions[i].freedomOfMovement+1);
			for(int j=0; j<actions[i].getDim(); ++j)
			{
				actionSize *= actions[i].gridSize;
				str << "Action." << i << "." << j;
				if(j != actions[i].getDim()-1)
					str << " ";
			}
			if(i != actions.size()-1)
				str << " ";
		}

		std::cout << "Delta Size = " << deltaSize << "\nAction Size = " << actionSize << "\n";

		qp.evaluate(str.str(), qp.getGlobalState(), output, true);


		//str = std::stringstream();
		str.str("");
		str << "begin ";
		str << "(new integer Delta.size) ";
		str << "(set Delta.size (integer " << deltaSize << ") ) ";
		str << "(new integer action-size) ";
		str << "(set action-size (integer " << actions.size() << ") ) ";
		str << "(new integer action-space-size) ";
		str << "(set action-space-size (integer " << actionSize << ") ) ";
		//str << "(new integer measure-result) ";

		// now construct Think operators
		int systemSize = 1;
		for(unsigned int i=0; i<actions.size(); ++i)
		{
			str << "(new integer Action." << i << ".center) ";
			str << "(set Action." << i << ".center (integer " << actions[i].gridSize/2 << ") ) ";

			/*for(unsigned int k=0; k<actions[i].getDim(); ++k)
			{
				str << "(new integer Action." << i << "." << 0 << ".direction." << k << ") ";
				str << "(set Action." << i << "." << 0 << ".direction." << k << " (integer " << 0 << ") ) ";
			}*/
			for(unsigned int j=0; j<actions[i].actions.size(); ++j)
			{
				for(unsigned int k=0; k<actions[i].actions[j].p.coord.size(); ++k)
				{
					str << "(new integer Action." << i << "." << j/*+1*/ << ".direction." << k << ") ";
					int dir = 0;
					if(actions[i].actions[j].p.coord[k] > 0)
						dir = 1;
					else if(actions[i].actions[j].p.coord[k] < 0)
						dir = -1;
					str << "(set Action." << i << "." << j/*+1*/ << ".direction." << k << " (integer " << dir << ") ) ";

					actions[i].actions[j].p.coord[k] += (actions[i].gridSize/2);
				}
			}
			//systemSize *= (actions[i].points.size()/*+1*/);
			//systemSize *= (2*actions[i].freedomOfMovement +1);
			for(int j=0; j<actions[i].getDim(); ++j)
			{
				systemSize *= actions[i].movementMap.size();//(2*actions[i].freedomOfMovement +1);
				systemSize *= actions[i].gridSize;
			}
		}
		systemSize *= memorySize;
		str << "(new integer system-size) ";
		str << "(set system-size (integer " << systemSize << ") )";

		//std::cout << str.str();
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		/*str = std::stringstream();
		str << "new system QDM-DeltaAction (integer " << actions.size()+1 << ") Delta ";
		for(unsigned int i=0; i<actions.size(); ++i)
		{
			for(int j=0; j<actions[i].getDim(); ++j)
			{
				str << "Action." << i << "." << j;
				if(j != actions[i].getDim()-1)
					str << " ";
			}
			if(i != actions.size()-1)
				str << " ";
		}
		
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);*/

		for(int i=0; i<(int)actions.size(); ++i)
		{
			for(int j=0; j<(int)actions[i].getDim(); ++j)
			{
				//str = std::stringstream();
				str.str("");
				str << "begin ";
				str << "(new matrix Tops." << i << "." << j << " (integer " << systemSize << ") (integer " << systemSize << ") ) ";
				str << "(set Tops." << i << "." << j << " (matrix (integer " << systemSize << ") (integer " << systemSize << ") zero) )";
				qp.evaluate(str.str(), qp.getGlobalState(), output, true);
			}
		}
		str.str("");
		str << "begin ";
		for(int i=0; i<(int)actions.size(); ++i)
		{
			str << "(new matrix ActionOpHistory." << i << " (integer " << systemSize << ") (integer " << systemSize << ") ) ";
			str << "(set ActionOpHistory." << i << " (matrix (integer " << systemSize << ") (integer " << systemSize << ") ident) ) ";
		}
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		//str = std::stringstream();
		str.str("");
		str << "begin ";
		str << "(new matrix Tops (integer " << systemSize << ") (integer " << systemSize << ") ) ";
		str << "(set Tops (matrix (integer " << systemSize << ") (integer " << systemSize << ") ident) )";
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);
		//shell();
		//str = std::stringstream();
		str.str("");
		str << "file QDM_Scripts/SetupTops.txt";
		qp.evaluate(str.str(), qp.getGlobalState(), output, false);
		//shell();

		//str = std::stringstream();
		str.str("");
		str << "begin ";
		str << "(new matrix psi.0 (integer " << systemSize << ") (integer 1) ) ";
		str << "(new matrix psi.init (integer " << systemSize << ") (integer 1) ) ";
		str << "(set psi.0 (matrix (integer " << systemSize << ") (integer 1) ident) )";
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		str.str("");
		str << "file QDM_Scripts/prepare.txt";
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		std::cout << "System Size = " << systemSize << "\n";
		//qp.shell();

		return true;
	}

	bool QDM::addSensor(Sensor& sensor)
	{
		sensor.data.randomize(deltaSize*memorySize);
		sensors.push_back(sensor);

		return true;
	}

	bool QDM::loadCenterOp(const std::string& filename)
	{
		std::ifstream f;
		f.open(filename.c_str(), std::ios_base::binary);
		if(!f.is_open())
			return false;

		int version;
		f.read((char*)&version, sizeof(int));
		if(version != 10) return false;

		Sensor sensor;
		int stringSize;
		f.read((char*)&stringSize, sizeof(int));
		sensor.name.resize(stringSize, ' ');
		for(int j=0; j<stringSize; ++j)
			sensor.name[j] = f.get();

		int vectorSize;
		f.read((char*)&vectorSize, sizeof(int));
		for(int j=0; j<vectorSize; ++j)
		{
			double value;
			f.read((char*)&value, sizeof(double));
			sensor.data.chi.push_back(value);
		}

		f.read((char*)&vectorSize, sizeof(int));
		for(int j=0; j<vectorSize; ++j)
		{
			double value;
			f.read((char*)&value, sizeof(double));
			sensor.data.phi.push_back(value);
		}

		f.read((char*)&vectorSize, sizeof(int));
		for(int j=0; j<vectorSize; ++j)
		{
			double value;
			f.read((char*)&value, sizeof(double));
			sensor.data.psi.push_back(value);
		}

		sensor.data.op.create(sensor.data.chi.size());
		sensor.data.refreshOp();
		centerOp = sensor;

		f.close();
		return true;
	}

	void QDM::createNewCenterOp()
	{
		centerOp.data.randomize(deltaSize*memorySize);
		centerOp.data.refreshOp();
	}

	bool QDM::setupTraining(int startPopulation, int _maxPopulation)
	{
		maxPopulation = _maxPopulation;
		sensorGA.clear();
		for(int i=0; i<startPopulation; ++i)
		{
			SensorDataGA data;
			data.sensors = sensors;
			for(unsigned int i=0; i<data.sensors.size(); ++i)
				data.sensors[i].data.randomize(deltaSize*memorySize);
			data.fitness = 0;
			sensorGA.push_back(data);
		}
		return true;
	}
	double QDM::calculateP(int actionSpaceIndex, int action, bool printPr)
	{
		if(actionSpaceIndex < 0 || actionSpaceIndex >= (int)actions.size())
			return 0;
		if(action < 0 || action > (int)actions[actionSpaceIndex].actions.size())
			return 0;

		int spaceSize0 = deltaSize*memorySize;
		for(int i=0; i<actionSpaceIndex; ++i)
		{
			for(int j=0; j<actions[i].getDim(); ++j)
				spaceSize0 *= actions[i].gridSize;
		}

		int spaceSize1 = 1;
		for(int i=0; i<actions[actionSpaceIndex].getDim(); ++i)
			spaceSize1 *= actions[actionSpaceIndex].gridSize;

		int spaceSize2 = 1;
		for(int i=actionSpaceIndex+1; i<(int)actions.size(); ++i)
		{
			for(int j=0; j<actions[i].getDim(); ++j)
				spaceSize2 *= actions[i].gridSize;
		}

		std::stringstream str;
		str << "begin ";
		str << "(new integer QDM-temp-Space.0.size) (new integer QDM-temp-Space.1.size) (new integer QDM-temp-Space.2.size) ";
		str << "(set QDM-temp-Space.0.size (integer " << spaceSize0 << ") ) ";
		str << "(set QDM-temp-Space.1.size (integer " << spaceSize1 << ") ) ";
		str << "(set QDM-temp-Space.2.size (integer " << spaceSize2 << ") ) ";
		str << "(new system QDM-temp-Space.0 (integer 3) QDM-temp-Space.0 QDM-temp-Space.1 QDM-temp-Space.2) ";
		str << "(new array QDM-temp-result double (integer " << spaceSize1 << ") ) ";
		str << "(measure_distribution (get psi.0) (get QDM-temp-Space.0) (integer 1) QDM-temp-result)";

		QProg::Object output;
		qp.evaluate(str.str(), qp.getGlobalState(), output, false);
		if(output.getType() != "vector")
			return 0;

		std::vector <double>* pr = QProg::ObjectHelper::getVector(output);

		if(printPr)
		{
			int counter = 0;
			for(int i=0; i<(int)pr->size(); ++i)
			{
				Point p2;
				p2.coord.push_back(i);
				std::cout << "p[" << i << "]:" << getAction(p2, actionSpaceIndex) << " = " << std::setprecision(4) << (*pr)[i] << "\t";
				++counter;
				if(counter >= 1)
				{
					counter = 0;
					std::cout << "\n";
				}
			}
		}

		Point p;
		p.coord.resize(actions[actionSpaceIndex].getDim());
		for(int i=0; i<actions[actionSpaceIndex].getDim(); ++i)
			p.coord[i] = 0;

		double sum = 0;
		for(int i=0; i<spaceSize1; ++i)
		{
			if(getAction(p, actionSpaceIndex) == action)
			{
				sum += (*pr)[i];
			}

			for(int j=actions[actionSpaceIndex].getDim()-1; j>=0; --j)
			{
				p.coord[j] ++;
				if(p.coord[j] == actions[actionSpaceIndex].gridSize)
					p.coord[j] = 0;
				else
					break;
			}
		}

		return sum;
	}

	void QDM::calculateDistribution(int actionSpaceIndex, std::vector <double>& actionDist, std::vector <double>& deltaDist)
	{
		if(actionSpaceIndex < 0 || actionSpaceIndex >= (int)actions.size())
			return;

		int spaceSize0 = deltaSize*memorySize;
		for(int i=0; i<actionSpaceIndex; ++i)
		{
			for(int j=0; j<actions[i].getDim(); ++j)
				spaceSize0 *= actions[i].gridSize;
		}

		int spaceSize1 = 1;
		for(int i=0; i<actions[actionSpaceIndex].getDim(); ++i)
			spaceSize1 *= actions[actionSpaceIndex].gridSize;

		int spaceSize2 = 1;
		for(int i=actionSpaceIndex+1; i<(int)actions.size(); ++i)
		{
			for(int j=0; j<actions[i].getDim(); ++j)
				spaceSize2 *= actions[i].gridSize;
		}

		//std::cout << "DEBUG: size1 = " << spaceSize0 << "\nsize2 = " << spaceSize1 << "\nsize3 = " << spaceSize2 << "\n\n";
		std::stringstream str;
		str << "begin ";
		str << "(new integer QDM-temp-Space.0.size) (new integer QDM-temp-Space.1.size) (new integer QDM-temp-Space.2.size) ";
		str << "(set QDM-temp-Space.0.size (integer " << spaceSize0 << ") ) ";
		str << "(set QDM-temp-Space.1.size (integer " << spaceSize1 << ") ) ";
		str << "(set QDM-temp-Space.2.size (integer " << spaceSize2 << ") ) ";
		str << "(new system QDM-temp-Space.0 (integer 3) QDM-temp-Space.0 QDM-temp-Space.1 QDM-temp-Space.2) ";
		str << "(new array QDM-temp-result double (integer " << spaceSize1 << ") ) ";
		str << "(measure_distribution (get psi.0) (get QDM-temp-Space.0) (integer 1) QDM-temp-result)";

		QProg::Object output;
		qp.evaluate(str.str(), qp.getGlobalState(), output, false);
		if(output.getType() != "vector")
			return;

		std::vector <double>* pr = QProg::ObjectHelper::getVector(output);
		actionDist = *pr;

		
		//str = std::stringstream();
		str.str("");
		str << "begin ";
		str << "(new integer QDM-temp-Space.0.size) (new integer QDM-temp-Space.1.size) (new integer QDM-temp-Space.2.size) ";
		str << "(set QDM-temp-Space.0.size (integer " << spaceSize0 << ") ) ";
		str << "(set QDM-temp-Space.1.size (integer " << spaceSize1 << ") ) ";
		str << "(set QDM-temp-Space.2.size (integer " << spaceSize2 << ") ) ";
		str << "(new system QDM-temp-Space.0 (integer 3) QDM-temp-Space.0 QDM-temp-Space.1 QDM-temp-Space.2) ";
		str << "(new array QDM-temp-result double (integer " << spaceSize0 << ") ) ";
		str << "(measure_distribution (get psi.0) (get QDM-temp-Space.0) (integer 0) QDM-temp-result)";

		QProg::Object output2;
		qp.evaluate(str.str(), qp.getGlobalState(), output2, false);
		if(output2.getType() != "vector")
			return;

		std::vector <double>* pr2 = QProg::ObjectHelper::getVector(output2);
		deltaDist = *pr2;
	}

	void QDM::setFitness(std::vector <TrainingSet>& set, SensorDataGA& sensor, int measureTime)
	{
		double epsilon = 0.1;
		sensor.fitness = 0;
		if(set.empty())
			return;

		sensor.ruleFitness.resize(set.size());
		/*sensor.individualFitness.resize(sensor.sensors.size());
		std::vector <double> avgAmount;
		for(unsigned int i=0; i<sensor.individualFitness.size(); ++i)
		{
			avgAmount.push_back(0);
			sensor.individualFitness[i] = 0;
		}*/

		algebra::mat temp;
		std::vector <Point> result;
		//for(int t=0; t<5; ++t)
		//{
		//prepare();
		for(unsigned int i=0; i<set.size(); ++i)
		{
			prepare();
			//measure(result);
			QProg::Object output;
			qp.getGlobalState().get("psi.0", output);
			algebra::mat *psi = QProg::ObjectHelper::getMatrix(output);
			int index = 0;
			for(int t=0; t<measureTime; ++t)
			{
				if(index < (int)set[i].states.sensors.size())
				{
					for(unsigned int j=0; j<set[i].states.sensors[index].size(); ++j)
					{
						for(unsigned int k=0; k<sensor.sensors.size(); ++k)
						{
							if(sensor.sensors[k].name == set[i].states.sensors[index][j])
								applySensor(&sensor.sensors[k].data.op);
						}
					}
				}
				/*for(unsigned int j=0; j<set[i].states.size(); ++j)
				{
					for(unsigned int k=0; k<sensor.sensors.size(); ++k)
					{
						if(sensor.sensors[k].name == set[i].states[j].name)
						{
							applySensor(&sensor.sensors[k].data.op);
						}
					}
				}*/
				think();

				++index;
				if(index >= (int)set[i].states.sensors.size() && set[i].states.mode == TrainingSensorSet::Repeat)
					index = 0;
			}
			double sum = 0;
			for(unsigned int j=0; j<set[i].distribution.size(); ++j)
			{
				double pr = calculateP(set[i].distribution[j].actionSpace, set[i].distribution[j].action) - set[i].distribution[j].p;
				pr *= pr;
				sum += pr;
			}

			sum += epsilon;
			sensor.fitness += (1.0/sum);

			sensor.ruleFitness[i] = (1.0/sum);

			/*for(unsigned int j=0; j<set[i].states.size(); ++j)
			{
				for(unsigned int k=0; k<sensor.sensors.size(); ++k)
				{
					if(sensor.sensors[k].name == set[i].states[j].name)
					{
						sensor.individualFitness[k] += (1.0/sum);
						avgAmount[k] += 1.0;
					}
				}
			}*/
			output.clear();
		}

		if(set.size() > 0)
			sensor.fitness /= (double)set.size();

		if(1==2&&set.size() > 1)
		{
			double sum = 0;
			for(int i=0; i<(int)set.size(); ++i)//(int)set.size()-1; ++i)
			{
				double e = ( (sensor.ruleFitness[i] - sensor.fitness)*(sensor.ruleFitness[i] - sensor.fitness) );
				if(e > 4)
					sum += e;
				//for(int j=i+1; j<(int)set.size(); ++j)
					//sum += (fabs(sensor.ruleFitness[i] - sensor.ruleFitness[j]));
			}

			sum /= (double)set.size();
			sum += epsilon;
			sensor.fitness /= sum;
		}

		/*for(int i=0; i<sensor.individualFitness.size(); ++i)
		{
			if(avgAmount[i] > 0.5)
				sensor.individualFitness[i] /= avgAmount[i];
		}*/
		//}

		//sensor.fitness /= 5.0;
	}

	void QDM::train(std::vector <TrainingSet>& set, int measureTime, bool recalculateFitness)
	{
		const int MAX_LIFETIME = 20;
		if(sensorGA.empty()) return;
		// first calculate fitness:
		std::list < SensorDataGA >::iterator Iter, Iter2;
		algebra::mat temp;
		if(recalculateFitness)
		{
			int populationSize = sensorGA.size(), counter = 0;
			for(Iter = sensorGA.begin(); Iter != sensorGA.end(); ++Iter)
			{
				std::cout << counter << "/" << populationSize << "     \r";
				++counter;
				SensorDataGA q = *Iter;
				setFitness(set, q, measureTime);
				*Iter = q;
			}

			sensorGA.sort();
		}
		std::cout << "DEBUG: max fitness = " << sensorGA.begin()->fitness << "\n";
		std::cout << "DEBUG: min fitness = " << sensorGA.rbegin()->fitness << "\n";

		// evolve:
		int startingSize = sensorGA.size();
		const int CROSSOVER_AMOUNT=4;
		for(int i=0; i<startingSize / CROSSOVER_AMOUNT + 1; ++i)
		{
#ifdef WIN32
			if(_kbhit()) break;
#endif
			std::cout << i << "/" << startingSize / CROSSOVER_AMOUNT+1 << "\n";
			int r = rand()%(startingSize / CROSSOVER_AMOUNT + 1);
			Iter = sensorGA.begin();
			for(int j=0; j<r; ++j)
			{
				++Iter;
				if(Iter == sensorGA.end())
				{
					Iter = sensorGA.begin();
					break;
				}
			}
			SensorDataGA child;
			r = rand()%startingSize;//(startingSize / CROSSOVER_AMOUNT + 1);
			Iter2 = sensorGA.begin();
			for(int j=0; j<r; ++j)
			{
				++Iter2;
				if(Iter2 == sensorGA.end())
				{
					Iter2 = sensorGA.begin();
					break;
				}
			}

			Iter->crossOver(*Iter2, child);
			setFitness(set, child, measureTime);
			//q.fitness(qnd, set);
			//std::cout << "\t" << q.getLastFitness() << "\n";
			sensorGA.push_back(child);
		}

		std::cout << "mutation...\n";
		Iter = sensorGA.begin();
		++Iter;
		for(; Iter != sensorGA.end(); ++Iter)
		{
#ifdef WIN32
			if(_kbhit())break;
#endif
			if(rand()%100 < 40)
			{
				//std::cout << Iter->fitness << " -> ";
				Iter->randomEvolve();
				SensorDataGA q = *Iter;
				setFitness(set, q, measureTime);
				//std::cout << q.fitness << "\n";
				*Iter = q;
				//Iter->fitness(qnd, set);
			}
		}

		//std::cout << "sort...\n";
		sensorGA.sort();
		for(int i=0; i<startingSize/CROSSOVER_AMOUNT +1; ++i)
		{
#ifdef WIN32
			if(_kbhit()) break;
#endif
			std::cout << i << "/" << startingSize / CROSSOVER_AMOUNT +1 << ": " << sensorGA.begin()->fitness << " -> ";
			SensorDataGA child;
			sensorGA.begin()->randomEvolve(child);
			setFitness(set, child, measureTime);
			std::cout << child.fitness << "\n";
			sensorGA.push_back(child);
		}
		sensorGA.sort();

		int endingSize = sensorGA.size();

		for(Iter = sensorGA.begin(); Iter != sensorGA.end(); ++Iter)
			Iter->lifetime ++;
		Iter = sensorGA.begin();
		for(int i=0; i<endingSize; ++i)
		{break;
			if(endingSize < 50) break;
			if(i > 9*endingSize/10 && Iter->lifetime >= MAX_LIFETIME)
			{
				sensorGA.erase(Iter);
				Iter = sensorGA.begin();
				--endingSize;
				i = -1;
				continue;
			}
			++Iter;
		}
		while(endingSize > maxPopulation)
		{
			sensorGA.pop_back();
			--endingSize;
		}

		//sensors = sensorGA.begin()->sensors;
		//double best = -1;
		//double best1=-1, best2=-1;
		/*std::cout << "Best = (";
		for(int i=0; i<sensors.size(); ++i)
		{
			best = -1;
			for(Iter=sensorGA.begin(); Iter != sensorGA.end(); ++Iter)
			{
				if(Iter->individualFitness[i] > best)
				{
					best = Iter->individualFitness[i];
				//	best1 = Iter->individualFitness[0];
				//	best2 = Iter->individualFitness[1];
					//sensors = Iter->sensors;
					sensors[i] = Iter->sensors[i];
				}
			}
			std::cout << best << ", ";// << best2;
		}
		std::cout << ")\n";*/

		sensors = sensorGA.begin()->sensors;

		std::cout << "DEBUG: final max fitness = " << sensorGA.begin()->fitness << "\n";
		for(int i=0; i<(int)sensorGA.begin()->ruleFitness.size(); ++i)
			std::cout << "\tRule " << i << " fitness: " << sensorGA.begin()->ruleFitness[i] << "\n";
		//std::cout << "DEBUG: min fitness = " << sensorGA.rbegin()->fitness << "\n";
	}

	void QDM::prepare()
	{
		QProg::Object output;
		std::stringstream str;
		//str << "file QDM_Scripts/prepare.txt";
		str << "set psi.0 (get psi.init)";
		qp.evaluate(str.str(), qp.getGlobalState(), output, false);
	}

	void QDM::applySensor(algebra::mat* op)
	{
		QProg::Object output;
		qp.getGlobalState().get("psi.0", output);
		algebra::mat *psi = QProg::ObjectHelper::getMatrix(output);

		algebra::mat temp, temp2;
		if(op->getRows() < actionSize*deltaSize*memorySize)
		{
			//std::cout << "Tensor Op\n";
			temp.create(actionSize);
			temp.ident();
			op->tensor(&temp, &temp2);
			temp2.multiply(*psi, temp);
		}
		else
		{
			op->multiply(psi, &temp);
		}
		*psi = temp;
		output.clear();
	}
	void QDM::applySensor(const std::string& name, bool useSecond)
	{
		if(useSecond)
			return;
		//QProg::Object output;
		//qp.getGlobalState().get("psi.0", output);
		//algebra::mat *psi = QProg::ObjectHelper::getMatrix(output), temp;

		for(unsigned int i=0; i<sensors.size(); ++i)
		{
			if(sensors[i].name == name)
			{
				if(useSecond)
				{
					algebra::mat tempOp;
					sensors[i].data.op.transpose(tempOp);
					applySensor(&tempOp);
				//	applySensor(&sensors[i].data2.op);
				}
				else
					applySensor(&sensors[i].data.op);
				//sensors[i].data.op.multiply(*psi, temp);
				//*psi = temp;
				break;
			}
		}
		//output.clear();
	}

	void QDM::applyInvSensor(const std::string& name)
	{
		for(unsigned int i=0; i<sensors.size(); ++i)
		{
			if(sensors[i].name == name)
			{
				algebra::mat tempOp;
				sensors[i].data.op.transpose(tempOp);
				applySensor(&tempOp);
				//sensors[i].data.op.multiply(*psi, temp);
				//*psi = temp;
				break;
			}
		}
		//output.clear();
	}
	void QDM::think()
	{
		QProg::Object output;
		std::stringstream str;
		str << "set psi.0 (mult (get Tops) (get psi.0) )";
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);
	}
	void QDM::measure(std::vector <Point>& result)
	{
		if(actions.size() == 0) return;
		QProg::Object output;
		result.clear();
		result.resize(actions.size());
		int index = actions.size();
		for(unsigned int i=0; i<actions.size(); ++i)
		{
			Point p;
			p.coord.resize(actions[i].getDim());
			for(int j=0; j<actions[i].getDim(); ++j)
			{
				std::stringstream str;
				str << "set psi.0 (measure (get psi.0) (get QDM) (integer " << index << ") Action." << i << "." << j << ".result)";
				qp.evaluate(str.str(), qp.getGlobalState(), output, false);

				//str = std::stringstream();
				str.str("");
				str << "Action." << i << "." << j << ".result";
				qp.getGlobalState().get(str.str(), output);
				p.coord[j] = *QProg::ObjectHelper::getInteger(output);
				output.clear();
				//std::cout << "p.coord[" << j << "] = " << p.coord[j] << "\n";
			}
			result[i] = p;
			++index;
		}

		// now to reset for next run (note that delta space may still be in a superposition)
		std::stringstream str;
		str << "file QDM_Scripts/reset.txt";
		//str << "file QDM_Scripts/prepare.txt";
		//qp.evaluate(str.str(), qp.getGlobalState(), output, false);
		reset(result);
	}

	void QDM::partialMeasure(Point& result, int actionSpaceIndex)
	{
		if(actionSpaceIndex < 0 || actionSpaceIndex >= (int)actions.size())
			return;

		if(actions.size() == 0) return;
		QProg::Object output;
		int index = 1;//actions.size()+actionSpaceIndex;
		for(int j=0; j<(int)actions.size(); ++j)
		{
			index += actions[j].getDim();
			if(j < actionSpaceIndex)
				index += actions[j].getDim();
		}

		Point p;
		p.coord.resize(actions[actionSpaceIndex].getDim());
		for(int j=0; j<actions[actionSpaceIndex].getDim(); ++j)
		{
			std::stringstream str;
			str << "set psi.0 (measure (get psi.0) (get QDM) (integer " << index << ") Action." << actionSpaceIndex << "." << j << ".result)";
			qp.evaluate(str.str(), qp.getGlobalState(), output, false);

			//str = std::stringstream();
			str.str("");
			str << "Action." << actionSpaceIndex << "." << j << ".result";
			qp.getGlobalState().get(str.str(), output);
			p.coord[j] = *QProg::ObjectHelper::getInteger(output);
			output.clear();
			++index;
			//std::cout << "p.coord[" << j << "] = " << p.coord[j] << "\n";
		}
		result = p;
		//qp.evaluate("qdm.debug.graph2D update1D", qp.getGlobalState(), output, false);
		reset(result, actionSpaceIndex);
		//prepare();
	}

	void QDM::reset(Point& result, int actionSpaceIndex)
	{
		int i = actionSpaceIndex;
		if(i < 0 || i >= (int)actions.size())
			return;

		algebra::mat temp, deltaIdent, final_op, temp2, *ket, *bra, *psi;
		deltaIdent.create(deltaSize);
		deltaIdent.ident();
		temp.create(actionSize);

		std::stringstream str;
		QProg::Object output;
		//qp.evaluate("file QDM_Scripts/prepare.txt", qp.getGlobalState(), output, false);
		//return;

		qp.getGlobalState().get("psi.0", output);
		psi = QProg::ObjectHelper::getMatrix(output);
		output.clear();

		temp.create(1);
		temp.ident();
		for(int j=0; j<actions[i].getDim(); ++j)
		{
			// set temp = |center><result|
			str.str("");
			str << "Action." << i << "." << j << ".ket." << actions[i].gridSize/2;
			qp.getGlobalState().get(str.str(), output);
			ket = QProg::ObjectHelper::getMatrix(output);
			output.clear();

			str.str("");
			str << "Action." << i << "." << j << ".bra." << result.coord[j];
			qp.getGlobalState().get(str.str(), output);
			bra = QProg::ObjectHelper::getMatrix(output);
			output.clear();

			ket->multiply(bra, &temp2);
			temp.tensor(&temp2, &final_op);
			temp = final_op;
		}

		// now apply "temp" to action space "i"
		int first=deltaSize*memorySize, last=1;
		for(int j=0; j<(int)actions.size(); ++j)
		{
			for(int k=0; k<(int)actions[j].getDim(); ++k)
			{
				if(j < i)
					first *= actions[j].gridSize;
				else if(j > i)
					last *= actions[j].gridSize;
			}
		}

		final_op.create(first);
		final_op.ident();
		final_op.tensor(&temp, &temp2);	// temp2 = I x op
		final_op.create(last);
		final_op.ident();
		temp2.tensor(&final_op, &temp);

		temp.multiply(psi, &final_op);
		*psi = final_op;
		if(!actions[actionSpaceIndex].resetDelta)
			return;
		//return;
		// now measure Delta space:
		int index = 1;
		for(int j=0; j<i; ++j)
			index += actions[j].getDim();

		Point result2 = result;
		for(int j=0; j<actions[i].getDim(); ++j)
		{
			str.str("");
			str << "set psi.0 (measure (get psi.0) (get QDM) (integer " << index << ") Action." << i << "." << j << ".result)";
			qp.evaluate(str.str(), qp.getGlobalState(), output, false);

			str.str("");
			str << "Action." << actionSpaceIndex << "." << j << ".result";
			qp.getGlobalState().get(str.str(), output);
			result2.coord[j] = *QProg::ObjectHelper::getInteger(output);
			output.clear();
			++index;
		}

		temp.create(1);
		temp.ident();
		for(int j=0; j<actions[i].getDim(); ++j)
		{
			// set temp = |center><result|
			str.str("");
			str << "Delta." << i << "." << j << ".ket.1";
			qp.getGlobalState().get(str.str(), output);
			ket = QProg::ObjectHelper::getMatrix(output);
			output.clear();

			str.str("");
			str << "Delta." << i << "." << j << ".bra." << result2.coord[j];
			qp.getGlobalState().get(str.str(), output);
			bra = QProg::ObjectHelper::getMatrix(output);
			output.clear();

			ket->multiply(bra, &temp2);
			temp.tensor(&temp2, &final_op);
			temp = final_op;
		}

		// now apply "temp" to Delta space "i"
		first=memorySize;
		last=actionSize;
		for(int j=0; j<(int)actions.size(); ++j)
		{
			for(int k=0; k<(int)actions[j].getDim(); ++k)
			{
				if(j < i)
					first *= actions[j].movementMap.size();
					//first *= (2*actions[j].freedomOfMovement+1);
				else if(j > i)
					last *= actions[j].movementMap.size();
					//last *= (2*actions[j].freedomOfMovement+1);
			}
		}

		std::cout << first << ", " << temp.getRows() << ", " << last << "     \r";

		final_op.create(first);
		final_op.ident();
		final_op.tensor(&temp, &temp2);	// temp2 = I x op
		final_op.create(last);
		final_op.ident();
		temp2.tensor(&final_op, &temp);

		temp.multiply(psi, &final_op);
		*psi = final_op;
	}

	void QDM::reset(std::vector <Point>& result)
	{
		// Action.i.j.result = result[i].coord[j]
		algebra::mat temp, deltaIdent, final_op, temp2, *psi;
		deltaIdent.create(deltaSize);
		deltaIdent.ident();
		temp.create(actionSize);

		std::stringstream str;
		QProg::Object output;
		qp.getGlobalState().get("psi.0", output);
		psi = QProg::ObjectHelper::getMatrix(output);
		output.clear();
		for(unsigned int i=0; i<actions.size(); ++i)
		{
			reset(result[i], i);
			/*temp.create(1);
			temp.ident();
			for(int j=0; j<actions[i].getDim(); ++j)
			{
				// set temp = |center><result|
				str = std::stringstream();
				str << "Action." << i << "." << j << ".ket." << actions[i].gridSize/2;
				qp.getGlobalState().get(str.str(), output);
				ket = QProg::ObjectHelper::getMatrix(output);
				output.clear();

				str = std::stringstream();
				str << "Action." << i << "." << j << ".bra." << result[i].coord[j];
				qp.getGlobalState().get(str.str(), output);
				bra = QProg::ObjectHelper::getMatrix(output);
				output.clear();

				ket->multiply(bra, &temp2);
				temp.tensor(&temp2, &final_op);
				temp = final_op;
			}

			// now apply "temp" to action space "i"
			int first=deltaSize, last=1;
			for(int j=0; j<actions.size(); ++j)
			{
				for(int k=0; k<actions[j].getDim(); ++k)
				{
					if(j < i)
						first *= actions[j].gridSize;
					else if(j > i)
						last *= actions[j].gridSize;
				}
			}

			final_op.create(first);
			final_op.ident();
			final_op.tensor(&temp, &temp2);	// temp2 = I x op
			final_op.create(last);
			final_op.ident();
			temp2.tensor(&final_op, &temp);

			temp.multiply(psi, &final_op);
			*psi = final_op;*/
		}
	}

	bool QDM::lookupAction(const std::string& name, int& actionSpaceIndex, int& actionIndex)
	{
		for(unsigned int i=0; i<actions.size(); ++i)
		{
			for(unsigned int j=0; j<actions[i].actions.size(); ++j)
			{
				if(actions[i].actions[j].name == name)
				{
					actionSpaceIndex = i;
					actionIndex = j;
					return true;
				}
			}
		}

		return false;
	}
	int QDM::getAction(Point& p, int index, Action* action)
	{
		if(index < 0 || index >= (int)actions.size() || actions[index].getDim() != p.coord.size())
			return 0;

		unsigned int smallest = 0;
		double dist = p.dist(actions[index].actions[0].p);
		for(unsigned int i=1; i<actions[index].actions.size(); ++i)
		{
			if(p.dist(actions[index].actions[i].p) < dist)
			{
				dist = p.dist(actions[index].actions[i].p);
				smallest = i;
			}
		}

		if(action)
			*action = actions[index].actions[smallest];
		return smallest;
	}

	int QDM::getMaxMeasureTime(int actionSpaceIndex)
	{
		if(actionSpaceIndex < 0 || actionSpaceIndex >= (int)actions.size())
			return 0;

		return actions[actionSpaceIndex].gridSize/2;
	}

	void CommandDebug::execute(wok::Text& text, QProg::State& state, QProg::Object& output, bool globalMode, bool copyState, QProg::QProg* qp)
	{
		QProg::Object mode_obj;
		qp->evaluate(text[1], state, mode_obj, globalMode, false);

		if(*QProg::ObjectHelper::getString(mode_obj) == "calcP")
		{
			QProg::Object actionIndex, action;
			qp->evaluate(text[2], state, actionIndex, globalMode, false);
			qp->evaluate(text[3], state, action, globalMode, false);

			QProg::ObjectHelper::newDouble(qdm->calculateP(*QProg::ObjectHelper::getInteger(actionIndex), *QProg::ObjectHelper::getInteger(action), false), output);
		}
		else if(*QProg::ObjectHelper::getString(mode_obj) == "listOp")
		{
			for(int i=0; i<(int)qdm->sensors.size(); ++i)
				std::cout << qdm->sensors[i].name << "[" << qdm->sensors[i].data.op.getRows() << "]\n";
		}
		else if(*QProg::ObjectHelper::getString(mode_obj) == "applyOp")
		{
			QProg::Object opName;
			qp->evaluate(text[2], state, opName, globalMode, false);

			qdm->applySensor(*QProg::ObjectHelper::getString(opName));
		}
		else if(*QProg::ObjectHelper::getString(mode_obj) == "applyInvOp")
		{
			QProg::Object opName;
			qp->evaluate(text[2], state, opName, globalMode, false);

			qdm->applyInvSensor(*QProg::ObjectHelper::getString(opName));
		}
		else if(*QProg::ObjectHelper::getString(mode_obj) == "pMeasure")
		{
			qdm::Point r;
			qdm->partialMeasure(r, 0);
			std::cout << "DEBUG: Action = " << qdm->getAction(r, 0) << "\n";
		}
		else if(*QProg::ObjectHelper::getString(mode_obj) == "think")
		{
			qdm->think();
		}
		else if(*QProg::ObjectHelper::getString(mode_obj) == "prepare")
		{
			qdm->prepare();
		}
	}

#ifdef USE_VISUAL_DEBUG
	void CommandGraph2D::execute(wok::Text& text, QProg::State& state, QProg::Object& output, bool globalMode, bool copyState, QProg::QProg* qp)
	{
		QProg::Object mode_obj;
		qp->evaluate(text[1], state, mode_obj, globalMode, false);

		if(*QProg::ObjectHelper::getString(mode_obj) == "connect")
		{
			graphClient->connect("127.0.0.1", 50024);
		}
		else if(*QProg::ObjectHelper::getString(mode_obj) == "create")
		{
			graphClient->createGraph3D(qdm->actions[0].gridSize, qdm->actions[0].gridSize);
			graphClient->newGraph("Action0");
			if(qdm->actions.size() == 2)
				graphClient->newGraph("Action1");
			graphClient->newGraph("Delta");
		}
		else if(*QProg::ObjectHelper::getString(mode_obj) == "update1D")
		{
			for(int i=0; i<(int)qdm->actions.size(); ++i)
			{
				std::vector <double> action, delta;
				qdm::Point p2;
				p2.coord.push_back(0);
				qdm->calculateDistribution(i, action, delta);

				std::vector <wok::NetRTGraphClient::Point> points;

				double width = 2.0;
				double dx = width / (double)action.size();
				double x = -width/2.0;
				std::stringstream g;
				g << "Action" << i;
				graphClient->clearGraph(g.str());
				for(int j=0; j<(int)action.size(); ++j)
				{
					wok::NetRTGraphClient::Point p;
					p.t = (float)x;
					p.y = (float)action[j];
					int a = qdm->getAction(p2, 0);
					p.c.r = 0;
					p.c.g = 0;
					p.c.b = 0;
					if(a == 0) p.c.r = 255;
					else if(a == 1) p.c.g = 255;
					else if(a == 2) {p.c.b = 255; p.c.g = 255; p.c.r = 255;}
					points.push_back(p);
					x += dx;
					p2.coord[0] ++;
				}
				graphClient->addPoints(g.str(), points);
				points.clear();

				if(i == 0)
				{
					dx = width / (double)delta.size();
					x = -width/2.0;
					graphClient->clearGraph("Delta");
					for(int i=0; i<(int)delta.size(); ++i)
					{
						wok::NetRTGraphClient::Point p;
						p.t = (float)x;
						p.y = (float)delta[i];
						points.push_back(p);
						x += dx;
					}
					graphClient->addPoints("Delta", points);
				}
			}
		}
		else if(*QProg::ObjectHelper::getString(mode_obj) == "update")
		{
			std::vector <double> dist, delta;
			qdm->calculateDistribution(0, dist, delta);

			std::vector <wok::NetRTGraphClient::Point3D> points;
			wok::NetRTGraphClient::Point3D p;
			int index = 0;
			qdm::Point qdmPoint;
			qdmPoint.coord.resize(2);
			for(int x=0; x<qdm->actions[0].gridSize; ++x)
			{
				for(int y=0; y<qdm->actions[0].gridSize; ++y)
				{
					p.x = x;
					p.y = y;
					p.z = 2.0f*(float)dist[index];
					qdmPoint.coord[0] = x;
					qdmPoint.coord[1] = y;
					int a = qdm->getAction(qdmPoint, 0);
					if(a == 0)
					{
						p.c.r = 1.0f;
						p.c.g = 1.0f;
						p.c.b = 1.0f;
					}
					else if(a == 1)
					{
						p.c.r = 1.0f;
						p.c.g = 0;
						p.c.b = 0;
					}
					else if(a == 2)
					{
						p.c.r = 0;
						p.c.g = 1.0f;
						p.c.b = 0;
					}
					else if(a == 3)
					{
						p.c.r = 0;
						p.c.g = 0;
						p.c.b = 1.0f;
					}
					else if(a == 4)
					{
						p.c.r = 0;
						p.c.g = 1.0f;
						p.c.b = 1.0f;
					}
					points.push_back(p);
					++index;
				}
			}

			graphClient->setPoints3D(points);
		}
	}
#endif

	/*
	Model 1
	bool QDM::addActionSpace(ActionSpace& A)
	{
		QProg::Object output;

		// create Delta space:
		std::stringstream str;
		str << "new space Delta." << actions.size() << " (integer " << A.action_degree.size() << ")";
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		// create Action space:
		str = std::stringstream();
		str << "begin ";
		int sum = 0;
		for(unsigned int i=0; i<A.action_degree.size(); ++i)
		{
			str << "(new integer action-deg." << actions.size() << "." << i << ") ";
			str << "(set action-deg." << actions.size() << "." << i << " (integer " << A.action_degree[i] << ") ) ";

			str << "(new integer action-index." << actions.size() << "." << i << ") ";
			str << "(set action-index." << actions.size() << "." << i << " (integer " << sum << ") ) ";
			sum += A.action_degree[i];
		}
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		str = std::stringstream();
		A.actionSize = sum;
		str << "new space Action." << actions.size() << " (integer " << sum << ")";
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		actions.push_back(A);
		return true;
	}

	bool QDM::finalizeSpace()
	{
		QProg::Object output;
		std::stringstream str;
		str << "new system QDM (integer " << 2*actions.size() << ") ";

		for(unsigned int i=0; i<actions.size(); ++i)
			str << "Delta." << i << " ";

		for(unsigned int i=0; i<actions.size(); ++i)
		{
			str << "Action." << i;
			if(i != actions.size()-1)
				str << " ";
		}

		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		str = std::stringstream();
		str << "begin ";
		str << "(new integer action-size) ";
		str << "(set action-size (integer " << actions.size() << ") ) ";

		// now construct Think operators
		int systemSize = 1;
		for(unsigned int i=0; i<actions.size(); ++i)
		{
			systemSize *= actions[i].action_degree.size() * actions[i].actionSize;
		}
		str << "(new integer system-size) ";
		str << "(set system-size (integer " << systemSize << ") )";

		std::cout << str.str();
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);
		for(int j=0; j<actions.size(); ++j)
		{
			str = std::stringstream();
			str << "begin ";
			str << "(new matrix Tops." << j << " (integer " << systemSize << ") (integer " << systemSize << ") ) ";
			str << "(set Tops." << j << " (matrix (integer " << systemSize << ") (integer " << systemSize << ") zero) )";
			qp.evaluate(str.str(), qp.getGlobalState(), output, true);
		}

		str = std::stringstream();
		str << "file QDM_Scripts/SetupTops.txt";
		qp.evaluate(str.str(), qp.getGlobalState(), output, false);

		str = std::stringstream();
		str << "begin ";
		str << "(new matrix psi.0 (integer " << systemSize << ") (integer 1) ) ";
		str << "(set psi.0 (matrix (integer " << systemSize << ") (integer 1) ident) )";
		qp.evaluate(str.str(), qp.getGlobalState(), output, true);

		qp.shell();

		return true;
	}
	*/
}
