/*
QDM.h
Written by Walter O. Krawec

Copyright (c) 2013 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef QDM_H_
#define QDM_H_

#include <vector>
#include <sstream>
#include <iostream>
#include <iomanip>

#include "../QProg/QProg.h"

#ifdef USE_VISUAL_DEBUG
#include <WOK/NetRTGraph.h>
#endif

namespace qdm
{
	struct Point
	{
		std::vector <int> coord;

		double dist(Point& p)
		{
			if(p.coord.size() != coord.size()) return 0;
			double sum=0;
			for(unsigned int i=0; i<coord.size(); ++i)
				sum += (coord[i] - p.coord[i])*(coord[i] - p.coord[i]);
			return sum;
		}
	};
	struct Action
	{
		std::string name;
		Point p;
	};
	struct ActionSpace
	{
		ActionSpace() : gridSize(1), resetDelta(true) {};
		//std::vector <Point> points;
		std::vector <Action> actions;
		std::vector <Point> movementMap;//degreeOfFreedom;
		int gridSize;//, freedomOfMovement;
		bool resetDelta;
		void addAction(Action& a)
		{
			if(actions.empty())
				actions.push_back(a);
			else if(a.p.coord.size() == getDim())
				actions.push_back(a);
		}
		int getDim()
		{
			if(actions.empty()) return 0;
			return actions[0].p.coord.size();
		}
		/*void addPoint(Point& p)
		{
			if(points.empty())
				points.push_back(p);
			else if(p.coord.size() == getDim())
				points.push_back(p);
		}
		int getDim()
		{
			if(points.empty()) return 0;
			return points[0].coord.size();
		}*/
	};
	struct SensorData
	{
		std::vector <double> phi, psi, chi;
		algebra::mat op;

		void randomize(int deltaSize)
		{
			int size1 = deltaSize*deltaSize;
			size1 = size1/2;
			phi.resize(size1);
			psi.resize(size1);
			chi.resize(deltaSize);

			for(int i=0; i<(int)phi.size(); ++i)
			{
				phi[i] = 3.1415*((double)(rand()%10000))/10000.0;
				psi[i] = 3.1415*((double)(rand()%10000))/10000.0;
				if(i < deltaSize)
					chi[i] = 3.1415*((double)(rand()%10000))/10000.0;
			}

			op.create(deltaSize);
			op.random_unitary(phi, psi, chi);
		}
		void refreshOp()
		{
			op.random_unitary(phi, psi, chi);
		}
	};
	struct Sensor
	{
		std::string name;
		SensorData data;//, data2;
	};

	struct TrainingDistribution
	{
		TrainingDistribution() : action(0), actionSpace(0), p(0.0) {};

		int action, actionSpace;
		double p;
	};

	struct TrainingSensorSet
	{
		enum Mode
		{
			Repeat=0,		// Repeat will train by repeating sensor[t]
			SingleUse		// SingleUse will use CenterOp after the last element of sensor[t]
		};

		TrainingSensorSet() : mode(Repeat) {};

		// sensor[i][j] is applied on think iteration "i" for all j
		std::vector < std::vector <std::string> > sensors;
		Mode mode;
	};
	// So this should be a collection of inputs (up to N where "N" is the number
	// of iterations before measuring); these inputs should create
	// a distribution over the actions specified by "distribution"
	struct TrainingSet
	{
		//std::vector <Sensor> states;
		//std::vector <TrainingSensorSet> states;
		TrainingSensorSet states;
		std::vector <TrainingDistribution> distribution;
	};

	struct SensorDataGA
	{
		SensorDataGA() : fitness(0), lifetime(0) {};
		std::vector<Sensor> sensors;
		std::vector <double> ruleFitness;
		double fitness;
		int lifetime;

		bool operator< (const SensorDataGA& other)
		{
			/*float fit1 = (individualFitness[0]+individualFitness[1])/2.0;
			float fit2 = (other.individualFitness[0]+other.individualFitness[1])/2.0;
			if(fit1 > fit2)
				return true;
			return false;*/
			if(fitness > other.fitness)
				return true;
			return false;
		}

		void randomEvolve()
		{
			int p = 20;
			for(unsigned int i=0; i<sensors.size(); ++i)
			{
				for(unsigned int j=0; j<sensors[i].data.phi.size(); ++j)
				{
					if(rand()%100 < p)
						sensors[i].data.phi[j] = 3.1415*(((double)(rand()%10000))/10000.0);
					if(rand()%100 < p)
						sensors[i].data.psi[j] = 3.1415*(((double)(rand()%10000))/10000.0);
					if(j < sensors[i].data.chi.size() && rand()%100 < p)
						sensors[i].data.chi[j] = 3.1415*(((double)(rand()%10000))/10000.0);

					/*if(rand()%100 < 25)
						sensors[i].data2.phi[j] = 3.1415*((double)(rand()%10000))/10000.0;
					if(rand()%100 < 25)
						sensors[i].data2.psi[j] = 3.1415*((double)(rand()%10000))/10000.0;
					if(j < sensors[i].data.chi.size() && rand()%100 < 25)
						sensors[i].data2.chi[j] = 3.1415*((double)(rand()%10000))/10000.0;*/
				}
				sensors[i].data.refreshOp();
				//sensors[i].data2.refreshOp();
			}
		}
		void randomEvolve(SensorDataGA& child)
		{
			child = *this;
			//child.sensors.resize(sensors.size());
			int p = 10;
			for(unsigned int i=0; i<sensors.size(); ++i)
			{
				//child.sensors[i].data.phi = sensors[i].data.phi;
				//child.sensors[i].data.psi = sensors[i].data.psi;
				//child.sensors[i].data.chi = sensors[i].data.chi;
				//child.sensors[i].data.op.create(sensors[i].data.op.getRows());
				for(unsigned int j=0; j<sensors[i].data.phi.size(); ++j)
				{
					if(rand()%100 < p)
						child.sensors[i].data.phi[j] = 3.1415*(((double)(rand()%10000))/10000.0);
					if(rand()%100 < p)
						child.sensors[i].data.psi[j] = 3.1415*(((double)(rand()%10000))/10000.0);
					if(j < sensors[i].data.chi.size() && rand()%100 < p)
						child.sensors[i].data.chi[j] = 3.1415*(((double)(rand()%10000))/10000.0);
				}
				child.sensors[i].data.refreshOp();
				//sensors[i].data2.refreshOp();
			}
		}

		void crossOver(SensorDataGA& other, SensorDataGA& child)
		{
			child = *this;
			int type = rand()%100;
			for(unsigned int i=0; i<sensors.size(); ++i)
			{
				int bound1 = rand()%sensors[i].data.phi.size();
				int bound2 = rand()%sensors[i].data.phi.size();
				for(int j=0; j<(int)sensors[i].data.phi.size(); ++j)
				{
					//if(j < sensors[i].data.phi.size()/2)
					if(j < bound1)
					{
						child.sensors[i].data.phi[j] = sensors[i].data.phi[j];
						child.sensors[i].data.psi[j] = sensors[i].data.psi[j];
					}
					else
					{
						child.sensors[i].data.phi[j] = other.sensors[i].data.phi[j];
						child.sensors[i].data.psi[j] = other.sensors[i].data.psi[j];
					}
				}

				int bound = rand()%sensors[i].data.chi.size();
				for(int j=0; j<(int)sensors[i].data.chi.size(); ++j)
				{
					if(j < bound)//sensors[i].data.chi.size()/2)
						child.sensors[i].data.chi[j] = sensors[i].data.chi[j];
					else
						child.sensors[i].data.chi[j] = other.sensors[i].data.chi[j];
				}
				child.sensors[i].data.refreshOp();

				// data2:
				/*bound1 = rand()%sensors[i].data2.phi.size();
				bound2 = rand()%sensors[i].data2.phi.size();
				for(unsigned int j=0; j<sensors[i].data2.phi.size(); ++j)
				{
					//if(j < sensors[i].data.phi.size()/2)
					if(j < bound1)
					{
						child.sensors[i].data2.phi[j] = sensors[i].data2.phi[j];
						child.sensors[i].data2.psi[j] = sensors[i].data2.psi[j];
					}
					else
					{
						child.sensors[i].data2.phi[j] = other.sensors[i].data2.phi[j];
						child.sensors[i].data2.psi[j] = other.sensors[i].data2.psi[j];
					}
				}

				bound = rand()%sensors[i].data2.chi.size();
				for(unsigned int j=0; j<sensors[i].data2.chi.size(); ++j)
				{
					if(j < bound)//sensors[i].data.chi.size()/2)
						child.sensors[i].data2.chi[j] = sensors[i].data2.chi[j];
					else
						child.sensors[i].data2.chi[j] = other.sensors[i].data2.chi[j];
				}
				child.sensors[i].data2.refreshOp();*/
			}
		}
	};

	class QDM
	{
	public:
		QDM() : memorySize(1), deltaSize(1), actionSize(1), maxPopulation(100) {};

		void setup(void* netRTGraph=NULL);

		bool addActionSpace(ActionSpace& A);
		bool finalizeSpace(int _memorySize);

		bool addSensor(Sensor& sensor);

		bool setupTraining(int startPopulation, int _maxPopulation);

		// this should calculate the probability of finding "psi.0" in a position
		// representing "action" in the space "actionSpaceIndex"
		double calculateP(int actionSpaceIndex, int action, bool printPr=false);

		void calculateDistribution(int actionSpaceIndex, std::vector <double>& actionDist, std::vector <double>& deltaDist);

		void setFitness(std::vector <TrainingSet>& set, SensorDataGA& sensor, int measureTime);
		void train(std::vector <TrainingSet>& set, int measureTime, bool recalculateFitness);

		void prepare();
		void applySensor(const std::string& name, bool useSecond=false);
		void applyInvSensor(const std::string& name);
		void applySensor(algebra::mat* op);
		void think();
		void measure(std::vector <Point>& result);
		void partialMeasure(Point& result, int actionSpaceIndex);
		void reset(std::vector <Point>& result);
		void reset(Point& result, int actionSpaceIndex);
		int getAction(Point& p, int index, Action* action=NULL);
		bool lookupAction(const std::string& name, int& actionSpaceIndex, int& actionIndex);

		int getMaxMeasureTime(int actionIndex);

		void shell() {qp.shell();};

		bool saveToFile(const std::string& filename);
		bool loadFromFile(const std::string& filename);
		bool loadFromFiles(const std::vector < std::string >& filenames);

		bool loadCenterOp(const std::string& filename);
		void createNewCenterOp();
	//private:
		QProg::QProg qp;
		std::vector <ActionSpace> actions;
		std::vector <Sensor> sensors;
		Sensor centerOp;
		std::list < SensorDataGA > sensorGA;

		int memorySize, deltaSize, actionSize;
		int maxPopulation;
	};

#ifdef USE_VISUAL_DEBUG
	class CommandGraph2D : public QProg::Command
	{
	public:
		CommandGraph2D(wok::NetRTGraphClient* _graphClient, QDM* _qdm) : QProg::Command("qdm.debug.graph2D"), graphClient(_graphClient), qdm(_qdm) {};
		virtual void execute(wok::Text& text, QProg::State& state, QProg::Object& output, bool globalMode, bool copyState, QProg::QProg* qp);
		virtual void printHelp()
		{
			std::cout << "Usage: qdm.debug.graph2D MODE:STRING [ARGS]\n";
			std::cout << "\tMODE=update: Will print prob. dist. over 2D action\n\t\tspace indexed by first argument\n";
			std::cout << "\tMODE=create: Sets the graph for the first time (call after\n\t\tfinalize action space)\n";
		}
	private:
		QDM* qdm;
		wok::NetRTGraphClient* graphClient;
	};
#endif

	class CommandDebug : public QProg::Command
	{
	public:
		CommandDebug(QDM* _qdm) : QProg::Command("qdm.debug"), qdm(_qdm) {};
		virtual void execute(wok::Text& text, QProg::State& state, QProg::Object& output, bool globalMode, bool copyState, QProg::QProg* qp);
		virtual void printHelp()
		{
			std::cout << "Usage: qdm.debug OPTION:STRING [ARGS]\n";
			std::cout << "\tMODE=calcP: Runs the 'calculateP' command\n\t\tARG1 = actionSpaceIndex:INTEGER\n\t\tARG2 = action:INTEGER\n";
			std::cout << "\tMODE=applyOp: Runs the 'applyOp' command\n\t\tARG1 = OP_NAME:STRING\n";
			std::cout << "\tMODE=applyInvOp: Runs the 'applyInvOp' command\n\t\tARG1 = OP_NAME:STRING\n";
			std::cout << "\tMODE=pMeasure: Runs the 'partialMeasure' command\n\t\tARG1 = actionSpaceIndex:INTEGER\n";
		}

	private:
		QDM* qdm;
	};

	/*
	Model 1: H = H(Delta) x H(Action)
	struct ActionSpace
	{
		ActionSpace() : actionSize(1) {};
		std::vector <int> action_degree;
		int actionSize;
	};

	class QDM
	{
	public:
		bool addActionSpace(ActionSpace& A);
		bool finalizeSpace();
	private:
		QProg::QProg qp;

		std::vector <ActionSpace> actions;
	};*/
}

#endif