/*
main.cpp
Written by Walter O. Krawec

Copyright (c) 2013 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

// NUM_BIRDS is the number of birds to create

#define NUM_BIRDS 8

// Leave this value alone (not really used now)
#define SPEED .2f

#include "particle.h"

#include <vector>
#include <time.h>

#include "Math3D.h"

#include "QBL/QBL.h"

wok::gl::glWindow glWindow;
wok::gl::Camera camera;
ParticleManager particles;
VisualFXManager effects;

struct Bird
{
	Bird() : x(0), y(0), rotation(0), targetRotation(0), QBL(NULL), call1(0), call2(0), updateIteration(1), velocity(1.0) {};
	Bird(float _x, float _y, float _r) : x(_x), y(_y), rotation(_r), targetRotation(_r), call1(0), call2(0), updateIteration(1), velocity(1.0) {};

	float x, y, rotation;
	float targetRotation;
	float velocity;

	qbl::QBL *QBL;
	int t;
	int call1, call2;

	int particleTimer;
	int updateCount;
	int updateIteration;

	void setup()
        {
		QBL = new qbl::QBL;
		QBL->evaluate("global (new qbl::QDM QDM)");
		QBL->evaluate("qbl::qdm::new (get QDM)");
		QBL->evaluate("file qbl_flock_setup.txt");
		QBL->evaluate("qbl::qdm::prepare (get QDM)");

		t = 0;

		particleTimer = rand()%10;

		updateCount = rand()%100;
		updateIteration = 3;
	}

	void cleanup()
	{
		if(QBL)
		{
			QBL->evaluate("qbl::qdm::free (get QDM)");
			delete QBL;
		}
	}
	void draw(ParticleManager& particles, int ourIndex)
	{
		++particleTimer;
		if(particleTimer > 10)
		{
			particleTimer = 0;
			particles.create(x,y);
		}
		float baseSize = 0.15f;
		//float height = .4f;
		baseSize = .35f;
		float height = 2.5*baseSize;
		math3D::CVector2D a,b,c;
				
		a.Construct(math3D::CMath3D::Deg2Rad(90+rotation), baseSize/2.0f);
		a.x += x;
		a.y += y;
				
		b.Construct(math3D::CMath3D::Deg2Rad(-90+rotation), baseSize/2.0f);
		b.x += x;
		b.y += y;
				
		c.Construct(math3D::CMath3D::Deg2Rad(0+rotation), height);
		c.x += x;
		c.y += y;

		switch(ourIndex)
		{
		case 0: glColor4f(1,1,1,1); break;
		case 1: glColor4f(0,0,1,1); break;
		case 2: glColor4f(0,1,0,1); break;
		case 3: glColor4f(1,0,0,1); break;
		default: glColor4f(1,1,1,1); break;
		}
		glBegin(GL_LINES);
		glVertex3f(a.x, a.y, 0); glVertex3f(b.x, b.y, 0);
		glVertex3f(b.x, b.y, 0); glVertex3f(c.x, c.y, 0);
		glVertex3f(c.x, c.y, 0); glVertex3f(a.x, a.y, 0);
		glEnd();
		glColor4f(1,1,1,1);
	}

	void updateQBL(bool turnLeft, bool turnRight, VisualFXManager& fx, std::vector <Bird>& birds, int ourIndex)
	{
		// uncomment this to use classical logic (see below on what to also comment)
		/*if(turnLeft)
			targetRotation = rotation + 20;
		if(turnRight)
			targetRotation = rotation - 20;*/

		++updateCount;
		if(updateCount % updateIteration != 0)
			return;
		wok::interp::Object o;
		QBL->evaluate("get QDM", o);
		call1 = 0;
		call2 = 0;
		if(call1 > 0)
			turnLeft = true;
		if(call2 > 0)
			turnRight = true;

		
		if(turnLeft == turnRight)
			QBL->evaluate("qbl::qdm::apply_op (get QDM) center");
		else if(turnLeft)
			QBL->evaluate("qbl::qdm::apply_op (get QDM) left");
		else if(turnRight)
			QBL->evaluate("qbl::qdm::apply_op (get QDM) right");

		QBL->evaluate("qbl::qdm::think (get QDM)");
		++t;
		if(t >= 4)		// 3 for flock
		{
			t = 0;
			wok::interp::Object output;
			
			//QBL->evaluate("qbl::qdm::decide (get QDM) 0", output);
			QBL->evaluate("qbl::qdm::decide_str QDM 0", output);
			//int action = *wok::interp::ObjectHelper::getInteger(output);
			std::string action(*wok::interp::ObjectHelper::getString(output));

			// uncomment this to use classical logic (make sure to uncomment other targetRotation call above)
			//return;
			if(action == "forward")
				targetRotation = rotation;
			else if(action == "left")
				targetRotation = rotation + 20;
			else if(action == "right")
				targetRotation = rotation - 20;
			/*if(action == 0)
				targetRotation = rotation;
			else if(action == 1)
				targetRotation = rotation + 20;
			else if(action == 2)
				targetRotation = rotation - 20;*/
		}
	}

	int rotationDistance(int r1, int r2)
	{
		r1 %= 360;
		if(r1 < 0)
			r1 = 360 + r1;
		r2 %= 360;
		if(r2 < 0)
			r2 = 360 + r2;

		if(abs(r2-r1) <= 180)
			return abs(r2-r1);
		else
		{
			if(r1 > 180)
				r1 = 360-r1;
			if(r2 > 180)
				r2 = 360-r2;
			return r1+r2;
		}
	}

	void update(std::vector <Bird>& birds, VisualFXManager& fx, int ourIndex, float fps)
	{
		fps = fps / 720.0f;
		const float shortDistance = 2.0f;
		float shortRotation = 0;
		int shortRotationCount = 0;

		float avgHeading = 0;
		float avgPositionX = 0, avgPositionY = 0;


		float turn = 0;
		bool turnLeft = false, turnRight = false;
		for(int i=0; i<(int)birds.size(); ++i)
		{
			if(i == ourIndex) continue;
			float d = math3D::CMath3D::FindDistanceSq2D(x,y, birds[i].x, birds[i].y);
			if(d <= shortDistance)
			{
				float theta = atan2f(birds[i].y - y, birds[i].x - x);
				theta = math3D::CMath3D::Rad2Deg(theta);

				int t = (int)theta;
				t %= 360;
				if(t < 0)
					t = 360 + t;

				int r = (int)rotation;
				if(r < 0) r = 360 + r;
				
				// avoid
				
				if(rotationDistance(r+1, t) < rotationDistance(r-1, t))
					turnRight = true;
				else if(rotationDistance(r+1, t) > rotationDistance(r-1, t))
					turnLeft = true;
			}

			avgHeading += birds[i].rotation;
			avgPositionX += birds[i].x;
			avgPositionY += birds[i].y;

		}

		avgHeading /= (float)(birds.size()-1);
		avgPositionX /= (float)(birds.size()-1);
		avgPositionY /= (float)(birds.size()-1);

		int t = (int)avgHeading;
		t %= 360;
		if(t < 0)
			t = 360 + t;
		int r = rotation;
		r = r % 360;
		if(r < 0) r = 360 + r;
		if(turnLeft == turnRight && rotationDistance(r, t) > 10)
		{
			// left is addition
			// right is subtraction
			if(rotationDistance(r+1, t) < rotationDistance(r-1, t))
				turnLeft = true;
			else if(rotationDistance(r+1, t) > rotationDistance(r-1, t))
				turnRight = true;
		}

		float positionRotation = math3D::CMath3D::Rad2Deg(atan2f(avgPositionY-y, avgPositionX-x));// - rotation;
		t = (int)positionRotation;
		t %= 360;
		if(t < 0)
			t = 360 + t;
		positionRotation = t;
		r = rotation;
		r = r % 360;
		if(r < 0) r = 360 + r;

		if(turnLeft == turnRight && rotationDistance(r, t) > 20)
		{
			if(rotationDistance(r+1, t) < rotationDistance(r-1, t))
				turnLeft = true;
			else if(rotationDistance(r+1, t) > rotationDistance(r-1, t))
				turnRight = true;
		}

		updateQBL(turnLeft, turnRight, fx, birds, ourIndex);

		float MAX_DIST = 1000;

		int rot = (int)targetRotation;
		rot %= 360;
		if(rot < 0) rot = 360 + rot;
		targetRotation = rot;

		float speed = 0.8f;

		r = rotation;
		r = r % 360;
		if(r < 0) r = 360 + r;
		rotation = r;

		if(rotationDistance(r+1, rot) < rotationDistance(r-1, rot))
			rotation += 1.0f/fps*speed;
		else if(rotationDistance(r+1, rot) > rotationDistance(r-1, rot))
			rotation -= 1.0f/fps*speed;

		math3D::CVector2D v;
		v.Construct(math3D::CMath3D::Deg2Rad(rotation), velocity*0.01f/fps);
		float d = math3D::CMath3D::FindDistanceSq2D(0,0,x,y);
		
		x += v.x;
		y += v.y;
	}
};

std::vector <Bird> birds;

void resetBirds()
{
	for(int i=0; i<(int)birds.size(); ++i)
	{
		birds[i].x = 0;
		birds[i].y = 0;
		birds[i].rotation = rand()%360;
		birds[i].targetRotation = birds[i].rotation;
	}
}

void setupBirds()
{
	for(int i=0; i<NUM_BIRDS; ++i)
	{
		Bird b;
		b.x = rand()%1000;
		b.x /= 1000.0f;
		b.y = rand()%1000;
		b.y /= 1000.0f;

		b.rotation = rand()%1000;
		b.rotation /= 1000.0f;
		b.targetRotation = b.rotation;

		b.x *= 4;
		b.x -= 2;
		b.y *= 4;
		b.y -= 2;

		b.setup();
		birds.push_back(b);
	}

	for(int i=0; i<50; ++i)
		particles.create(camera, Particle::Speck);
}
void cleanupBirds()
{
	for(int i=0; i<(int)birds.size(); ++i)
		birds[i].cleanup();
}
bool paused = true;
float fps = 1;
int fpsCounter = 1;
wok::Timer fpsTimer;
void draw()
{
	++fpsCounter;
	if(fpsTimer.done())
	{
		fps = (float)fpsCounter;
		fps *= 4.0f;
		fpsTimer.start();
		fpsCounter = 0;
	}
	//float fps = SPEED;
	glWindow.beginDraw();
	camera.select();
	particles.draw(camera);
	float avgX=0, avgY=0;
	float maxDist = math3D::CMath3D::FindDistanceSq2D(-camera.xPos(), -camera.yPos(),birds[0].x, birds[0].y);
	for(int i=0; i<(int)birds.size(); ++i)
	{
		if(paused)
			break;
		birds[i].update(birds, effects, i, fps);
		birds[i].draw(particles, i);
		avgX += birds[i].x;
		avgY += birds[i].y;

		float d = math3D::CMath3D::FindDistanceSq2D(-camera.xPos(),-camera.yPos(),birds[i].x, birds[i].y);
		if(d > maxDist)
			maxDist = d;
	}

	effects.draw();

	maxDist /= 2.0f;
	//maxDist = sqrtf(maxDist);
	if(maxDist < 5)
		maxDist = 5;
	if(maxDist > 50)
		maxDist = 50;
	avgX /= (float)birds.size();
	avgY /= (float)birds.size();

	camera.xPos() = -avgX;
	camera.yPos() = -avgY;
	//camera.zPos() = -maxDist;
	glWindow.endDraw();
}

#ifdef WIN32
LRESULT CALLBACK WndProc(	HWND	hWnd,			// Handle For This Window
							UINT	uMsg,			// Message For This Window
							WPARAM	wParam,			// Additional Message Information
							LPARAM	lParam)			// Additional Message Information
{
	switch (uMsg)									// Check For Windows Messages
	{
		case WM_SYSCOMMAND:
		{
			switch (wParam)
			{
				case SC_SCREENSAVE:
				case SC_MONITORPOWER:
					return 0;
			}
			break;
		}
	}

	glWindow.wndProc(hWnd, uMsg, wParam, lParam);

	// Pass All Unhandled Messages To DefWindowProc
	return DefWindowProc(hWnd,uMsg,wParam,lParam);
}

int WINAPI WinMain(	HINSTANCE	hInstance,			// Instance
					HINSTANCE	hPrevInstance,		// Previous Instance
					LPSTR		lpCmdLine,			// Command Line Parameters
					int			nCmdShow)			// Window Show State
{
	MSG		msg;									// Windows Message Structure
	BOOL	done=FALSE;								// Bool Variable To Exit Loop

	// Ask The User Which Screen Mode They Prefer
	srand(time(NULL));
	// Create Our OpenGL Window
	int width = 800;
	int height = 600;
	if (!glWindow.createWindow(WndProc, "Open GL",width,height,16,false))
	{
		return 0;									// Quit If Window Was Not Created
	}

	camera.zPos() = -30;
	setupBirds();

	fpsTimer.start(250);
	while(!done)									// Loop That Runs While done=FALSE
	{
		if (PeekMessage(&msg,NULL,0,0,PM_REMOVE))	// Is There A Message Waiting?
		{
			if (msg.message==WM_QUIT)				// Have We Received A Quit Message?
			{
				done=TRUE;							// If So done=TRUE
			}
			else									// If Not, Deal With Window Messages
			{
				TranslateMessage(&msg);				// Translate The Message
				DispatchMessage(&msg);				// Dispatch The Message
			}
		}
		else										// If There Are No Messages
		{
			// Draw The Scene.  Watch For ESC Key And Quit Messages From DrawGLScene()
			if(glWindow.getKey('P', true))
				paused = !paused;

			draw();
			if (glWindow.getKey(VK_ESCAPE))	// Active?  Was There A Quit Received?
			{
				done=TRUE;							// ESC or DrawGLScene Signalled A Quit
			}
			else if(glWindow.getKey('R', true))
				resetBirds();
			if(glWindow.getKey('Z'))
				camera.zPos() += 0.2f;
			if(glWindow.getKey('X'))
				camera.zPos() -= 0.2f;
		}
	}

	// Shutdown
	cleanupBirds();
	glWindow.destroyWindow();
	return (msg.wParam);							// Exit The Program
}
#else
void ReSizeGLScene(GLsizei width, GLsizei height)
{
	glWindow.resizeScreen(width, height);
}

void processNormalKeys(unsigned char key, int x, int y)
{
	glWindow.setKey(key, true);
}
void idleLoop()
{
  if(glWindow.getKey('p', true))
    paused = !paused;
  
  if(glWindow.getKey('r', true))
    resetBirds();

  if(glWindow.getKey('z', true))
    camera.zPos() += 0.2f;
  if(glWindow.getKey('x', true))
    camera.zPos() -= 0.2f;

  glutPostRedisplay();
}
int main(int argc, char** argv)
{
	srand(time(NULL));

	camera.zPos() = -30;
	setupBirds();

	fpsTimer.start(250);

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutInitWindowSize(800, 600);
	glutInitWindowPosition(100,100);
	glutCreateWindow(argv[0]);

	glWindow.initGL();

	glutDisplayFunc(draw);
	glutReshapeFunc(ReSizeGLScene);
	glutKeyboardFunc(processNormalKeys);

	glutIdleFunc(idleLoop);
	glutMainLoop();
	return 0;
}
#endif
